/*	File:	mm.c 
	Author:	Alex Fedosov
	Date:	September 21, 2000
	Description:	Memory Manager implementation for DMZ Project Core API
*/

/* === Includes === */
#include <stdlib.h>
#include "dmzcore.h"

/* === Constants === */

/* amount of memory intially allocated */
#define DMZ_MM_INITIAL_SIZE 4194304
#define DMZ_MM_FREE_LIST_SIZE 128


/* === Type Definitions === */

/* node in the array of free blocks */
typedef struct dmz_mm_free_node
{
	char *mem;
	unsigned int size;
} MM_FREE_NODE;

/* === Function Prototypes === */

/* initialize the MM by getting a big chunk of memory from the heap */
void* dmz_mm_init(size_t size);

/* === Global Variables === */
char *dmz_memory = NULL;
MM_FREE_NODE free_list[DMZ_MM_FREE_LIST_SIZE];


/* === Implementation === */

void* dmz_malloc(size_t size)
{
	void *ptr = NULL;
	int i;

	/* sanity check */
	dmz_assert(size);

	/* allocate memory if we haven't yet */
	if (!dmz_memory)
	{
		dmz_memory = dmz_mm_init(DMZ_MM_INITIAL_SIZE);

		/* initialize the free list */
		for (i = 0; i < DMZ_MM_FREE_LIST_SIZE; i++)
		{
			free_list[i].size = 0;
			free_list[i].mem = NULL;
		}

		/* now add that chunk to the free list */
		free_list[0].mem = dmz_memory;
		free_list[0].size = DMZ_MM_INITIAL_SIZE;
	}

	/* go through the list and find first matching chunk of available memory */
	for (i = 0; i < DMZ_MM_FREE_LIST_SIZE; i++)
		if (free_list[i].size >= size) /* chunk big enough, bite off a piece */
		{
			ptr = free_list[i].mem;
			free_list[i].size -= size;
			free_list[i].mem += size;
			break;
		}

	/* did we get something? */
	if (ptr == NULL)
		dmz_logerr("[core:mm:dmz_malloc] no available memory");

	return ptr;
}

void dmz_free(void *ptr, size_t size)
{
	int i;

	dmz_assert(size);
	dmz_assert(ptr);

	/* return this piece to the free list */
	for (i = 0; i < DMZ_MM_FREE_LIST_SIZE; i++)
		/* look for an available slot */
		if (free_list[i].size == 0)
		{
			free_list[i].size = size;
			free_list[i].mem = ptr;
			ptr = NULL;
		}

	/* make sure we had an available slot */
	if (ptr != NULL)
		dmz_logerr("[core:mm:dmz_free] free list full! leaking like a sieve");
}

/* initialize the MM by getting a big chunk of memory from the system */
void* dmz_mm_init(size_t size)
{
	void *ptr = malloc(size);

	if (ptr == NULL)
	{
		/* not enough memory available in the system.
		   complain and leave. */
		dmz_die("[core:mm:dmz_mm_init] not enough memory available!");
	}
	
	return ptr;
}
