/*	File:	blit.cpp
	Author:	Alex Fedosov
	Date:	March 14, 2001
	Description:	surface-to-surface blitting code
*/

#include <windows.h>
#include <ddraw.h>

#include <stdio.h>
#include <string.h>

#include "dmzgraph.h"
#include "dmzgraph_internal.h"

int dmz_graph_blit(int imgID, int src_x, int src_y, int src_w, int src_h,
				   int dst_x, int dst_y, int transparent)
{
	return dmz_graph_virblit(imgID, 0, src_x, src_y, src_w, src_h,
		dst_x, dst_y, transparent);
}

int dmz_graph_virblit(int srcID, int dstID, int src_x, int src_y, int src_w, int src_h,
					  int dst_x, int dst_y, int transparent)
{
	HRESULT r;
	DDSURFACEDESC srcdesc, dstdesc;
	RECT srect, drect;
	int mult = display_depth >> 3;

	if (!init_done) return DMZ_GRAPH_NOT_INIT;

	if (srcID < 0 || srcID >= MAX_IMAGES) return DMZ_GRAPH_INVALID_PARAM;
	if (dstID < 0 || dstID >= MAX_IMAGES) return DMZ_GRAPH_INVALID_PARAM;
	if (images[srcID] == NULL || images[dstID] == NULL) return DMZ_GRAPH_INVALID_PARAM;

	ZeroMemory(&srcdesc, sizeof(srcdesc));
	srcdesc.dwSize = sizeof(srcdesc);

	ZeroMemory(&dstdesc, sizeof(dstdesc));
	dstdesc.dwSize = sizeof(dstdesc);

	srect.left = src_x;
	srect.top = src_y;
	srect.right = src_x + src_w;
	srect.bottom = src_y + src_h;

	drect.left = dst_x;
	drect.top = dst_y;
	drect.right = dst_x + src_w;
	drect.bottom = dst_y + src_h;

	r = images[dstID]->Lock(&drect, &dstdesc, DDLOCK_WAIT | DDLOCK_WRITEONLY, NULL);
	if (r != DD_OK)
	{
		last_error = 0 - (r & 0xFFFF);
		return DMZ_GRAPH_DDRAW_ERROR;
	}

	r = images[srcID]->Lock(&srect, &srcdesc, DDLOCK_READONLY, NULL);
	if (r != DD_OK)
	{
		images[dstID]->Unlock(0);
		last_error = 0 - (r & 0xFFFF);
		return DMZ_GRAPH_DDRAW_ERROR;
	}

	BYTE *lpbDest = (BYTE *) dstdesc.lpSurface;
	BYTE *lpbSrc = (BYTE *) srcdesc.lpSurface;

	/* if transparency not specified, can use optimized blitting */
	if (transparent < 0)
	{
		int linewidth = src_w * mult;

		for (int i = 0; i < src_h; i++)
		{
			memcpy(lpbDest, lpbSrc, linewidth);

			lpbDest += dstdesc.lPitch;
			lpbSrc += srcdesc.lPitch;
		}
	}

	else
	{
		switch (display_depth)
		{
		case 16:
			__blit16(lpbSrc, lpbDest, src_w, src_h,
				srcdesc.lPitch, dstdesc.lPitch,	transparent);
			break;
		case 24:
			__blit24(lpbSrc, lpbDest, src_w, src_h,
				srcdesc.lPitch, dstdesc.lPitch,	transparent);
			break;
		case 32:
			__blit32(lpbSrc, lpbDest, src_w, src_h,
				srcdesc.lPitch, dstdesc.lPitch,	transparent);
			break;
		}
/*		for (int i = 0; i < src_h; i++)
		{
			for (int j = 0; j < src_w; j++)
			{
				DWORD *lpdwSrc = (DWORD *) lpbSrc;
				DWORD *lpdwDest = (DWORD *) lpbDest;

				if ((lpdwSrc[j] & 0xFFFFFF) != (unsigned) transparent)
					lpdwDest[j] = lpdwSrc[j];
			}

			lpbDest += dstdesc.lPitch;
			lpbSrc += srcdesc.lPitch;
		}
*/	}

	images[srcID]->Unlock(0);
	images[dstID]->Unlock(0);

	return DMZ_GRAPH_SUCCESS;
}

inline void __blit32(BYTE *lpbSrc, BYTE *lpbDest, int src_w, int src_h,
					  LONG srcPitch, LONG dstPitch, int transparent)
{
	for (int i = 0; i < src_h; i++)
	{
		for (int j = 0; j < src_w; j++)
		{
			DWORD *lpdwSrc = (DWORD *) lpbSrc;
			DWORD *lpdwDest = (DWORD *) lpbDest;

			if ((lpdwSrc[j] & 0xFFFFFF) != (unsigned) transparent)
				lpdwDest[j] = lpdwSrc[j];
		}

		lpbDest += dstPitch;
		lpbSrc += srcPitch;
	}
}

inline void __blit24(BYTE *lpbSrc, BYTE *lpbDest, int src_w, int src_h,
					  LONG srcPitch, LONG dstPitch, int transparent)
{
	for (int i = 0; i < src_h; i++)
	{
		RGBTRIPLE *rgbSrc = (RGBTRIPLE *) lpbSrc;
		RGBTRIPLE *rgbDest = (RGBTRIPLE *) lpbDest;

		for (int j = 0; j < src_w; j++, rgbSrc++, rgbDest++)
		{
			DWORD r = rgbSrc->rgbtRed << loRedBit;
			DWORD g = rgbSrc->rgbtGreen << loGrnBit;
			DWORD b = rgbSrc->rgbtBlue << loBluBit;
			DWORD pixel = r | g | b;;

			if ((pixel & 0xFFFFFF) != (unsigned) transparent)
				*rgbDest = *rgbSrc;
		}

		lpbDest += dstPitch;
		lpbSrc += srcPitch;
	}
}

inline void __blit16(BYTE *lpbSrc, BYTE *lpbDest, int src_w, int src_h,
					  LONG srcPitch, LONG dstPitch, int transparent)
{
	for (int i = 0; i < src_h; i++)
	{
		for (int j = 0; j < src_w; j++)
		{
			WORD *lpwSrc = (WORD *) lpbSrc;
			WORD *lpwDest = (WORD *) lpbDest;
			WORD trans = (WORD) (transparent & 0xFFFF);

			if ((lpwSrc[j] & 0xFFFF) != trans)
				lpwDest[j] = lpwSrc[j];
		}

		lpbDest += dstPitch;
		lpbSrc += srcPitch;
	}
}

int dmz_graph_setpixel(int dstID, int dst_x, int dst_y, unsigned int color)
{
	HRESULT r;
	DDSURFACEDESC dstdesc;
	int mult = display_depth >> 3;

	if (!init_done) return DMZ_GRAPH_NOT_INIT;

	if (dstID < 0 || dstID >= MAX_IMAGES) return DMZ_GRAPH_INVALID_PARAM;
	if (images[dstID] == NULL) return DMZ_GRAPH_INVALID_PARAM;

	ZeroMemory(&dstdesc, sizeof(dstdesc));
	dstdesc.dwSize = sizeof(dstdesc);

	r = images[dstID]->Lock(NULL, &dstdesc, DDLOCK_WAIT | DDLOCK_WRITEONLY, NULL);
	if (r != DD_OK)
	{
		last_error = 0 - (r & 0xFFFF);
		return DMZ_GRAPH_DDRAW_ERROR;
	}

	DWORD *lpdwDest = (DWORD *) ((BYTE *) dstdesc.lpSurface +
		dst_y * dstdesc.lPitch + dst_x * mult);
	WORD *lpwDest = (WORD *) ((BYTE *) dstdesc.lpSurface +
		dst_y * dstdesc.lPitch + dst_x * mult);

	switch (display_depth)
	{
	case 16:
		*lpwDest = (WORD) color;
		break;
	case 24:
	case 32:
		*lpdwDest = (DWORD) color;
		break;
	}

	images[dstID]->Unlock(0);

	return DMZ_GRAPH_SUCCESS;
}
